import asyncio
from typing import Callable

class Tweens:
    """A class for creating background functions that will change some property over time.
    """
    tweens: list = []

    def __init__(self):
        self.tweens = []

    def add(self, target: any,
            property: str,
            fr: float,
            to: float,
            repeat: int = 0,
            duration: int = 1000,
            delay: int = 0,
            after_tween: Callable | None = None,
            update_tween: Callable | None = None):
        """Add new tween.

        Args:
            target          Object that will be changed.
            property        Property that will be changed. Must be float.
            fr              Value to start from.
            to              Value to end on.
            repeat          Amount of times to repeat.
            duration        Duration of the change.
            after_tween     Function that will be called after the tween is ended.
            update_tween    Function that will be called every tick while the tween is on going.
        """
        loop = asyncio.get_event_loop()
        loop.create_task(self.start_tween(target, property, fr, to, repeat, duration, delay, after_tween, update_tween))
        
    async def start_tween(self, target: any,
            property: str | None = None,
            fr: float | None = None,
            to: float | None = None,
            repeat: int = 0,
            duration: int = 1000,
            delay: int = 0,
            after_tween: Callable | None = None,
            update_tween: Callable | None = None):
        await asyncio.sleep(delay)
        cnt = (1, repeat + 1)[repeat != -1]
        while cnt > 0:
            for i in range(60 * duration):
                if update_tween:
                    update_tween()
                else:
                    setattr(target, property, fr + (to - fr) * (i / (60 * duration)))
                await asyncio.sleep(1 / 60)

            if repeat != -1:
                cnt -= 1

        if after_tween:
            after_tween()
 



from typing import Callable
import pygame

from ..vmath import Vector2d
from ..game_object import Component, CoordinatesComponent, SurfaceComponent

class LabelComponent(Component):
    """A component that represents one line of text.

    Attributes:
        font    Pygame font object.
        text    Content of thsis label.
        color   Color of the text.
    """

    font: pygame.font.Font
    text: str
    color: tuple[int, int, int]

    def __init__(self, text: str, color: tuple[int, int, int], font_name="Arial", size=50):
        self.font = pygame.font.SysFont(font_name, size)
        self.text = text
        self.color = color
        
    def draw(self):
        text = self.font.render(self.text, True, self.color)
        text = pygame.transform.scale(text, (text.get_width(), text.get_height()))
        self.game_object.surface.pg_surf.blit(
            text, 
            text.get_rect(center=(self.game_object.surface.pg_surf.get_width() // 2, 
                                  self.game_object.surface.pg_surf.get_height() // 2))
        )

class ShapeComponent(Component):
    """Generic class for the components that represent geometric shapes.

    color   Fill color of the shape.
    """
    def __init__(self, color: tuple[int, int, int]):
        self.color = color

    def draw(self):
        pass

    def interception(self, position: Vector2d) -> bool:
        """Checks if position is inside of the shape.
        
        Args:
            position: Position to check.
        
        Returns:
            True if the position is inside of the shape, False otherwise.
        """
        return False

class RectShapeComponent(ShapeComponent):

    def __init__(self, color: tuple[int, int, int], size: Vector2d):
        super().__init__(color)
        self.size = size

    def draw(self):
        pygame.draw.rect(self.game_object.surface.pg_surf, self.color, 
                        (self.game_object.surface.pg_surf.get_width() // 2 - self.size.x // 2, 
                                self.game_object.surface.pg_surf.get_height() // 2 - self.size.y // 2, 
                                self.size.x, self.size.y))

    def interception(self, position: Vector2d) -> bool:
        center = self.game_object.transform.position
        temp = (center - position).operation(self.size, lambda a, b: -b/2 <= a <= b/2)
        return bool(temp.x) and bool(temp.y)
    

class CircleShapeComponent(ShapeComponent):
    def __init__(self, color: tuple[int, int, int], radius: float):
        super().__init__(color)
        self.radius = radius

    def draw(self):
        pygame.draw.circle(self.game_object.surface.pg_surf, self.color, 
                        (self.game_object.surface.pg_surf.get_width() // 2, 
                            self.game_object.surface.pg_surf.get_height() // 2), 
                        self.radius)

    def interception(self, position: Vector2d) -> bool:
        center = self.game_object.transform.position
        return (center - position).mod() <= self.radius


class ButtonComponent(Component):
    """Clickable component with its own shape.

    cmd     Function that will be called when button is clicked.
    """

    cmd: Callable

    def __init__(self, cmd: Callable, *args):
        self.cmd = cmd
        self.args = args

    def update(self):
        if pygame.mouse.get_pressed(3)[0]: # 0 - Left Mouse Button
            pos = self.game_object.get_component(CoordinatesComponent).get_cursor_coords()
            if self.game_object.get_component(ShapeComponent).interception(pos):
                self.cmd(self.args)

    def draw(self):
        self.game_object.get_component(ShapeComponent).draw()



from .angle import *
from .vector import *
from .modx import *
from .others import *
from .directions import *

__all__ = [
    'Angle',
    'Vector2d',
    'Modx',
    'complex_multiply',
    'Direction',
    'Directions',
    'is_in_box',
]


from typing import TypeVar
from .vmath import Vector2d, Angle

import pygame as pg

class Component:
    """Base class for all components.

    game_object     Associated GameObject.
    """
    game_object: "GameObject"

    class ComponentData:
        component_type: type

        def __init__(self, ctp: type):
            self.component_type = ctp

    def init(self, go: "GameObject"):
        self.game_object = go

    def draw(self):
        pass

    def update(self):
        pass

    def destroy(self):
        pass

    def get_data(self) -> ComponentData:
        return self.ComponentData(Component)


class Transform:
    """Contains all of the necessary data about GameObject's location.

    game_object     Associated game object.
    position        Position of the center of the game_object.
    angle           Angle for which game_object is rotated.

    """
    game_object: "GameObject"
    position: Vector2d  # position of CENTER
    angle: Angle
    scale: float

    def __init__(self, obj: "GameObject"):
        self.game_object = obj
        self.position = Vector2d(0, 0)
        self.angle = Angle()
        self.scale = 1
        self.childs = []

    def rotate(self, angle: Angle):
        self.angle += angle
        # TODO rotate all of the childs relative to center of this transform.

    def add_child(self, child: "Transform"):
        self.childs.append(child)

class SurfaceComponent(Component):
    """Necessary component to handle everything with game_object's display

    pg_surf     Associated pygame.Surface 
    """
    pg_surf: pg.Surface

    def __init__(self, size: Vector2d):
        self.pg_surf = pg.Surface(size.as_tuple(), pg.SRCALPHA, 32)

    def blit(self):
        surf = pg.display.get_surface()
        if self.game_object.parent:
            surf = self.game_object.parent.surface.pg_surf
        
        pos = self.game_object.transform.position
        if not self.game_object.parent:
            pos -= GameObject.get_by_tag("camera").transform.position

        surf.blit(self.pg_surf, (pos + Vector2d.from_tuple(surf.get_size()) / 2 - Vector2d.from_tuple(self.pg_surf.get_size()) / 2).as_tuple())


class CoordinatesComponent(Component):
    """Necessary component to get global object's coordinates.
    """

    def get_cursor_coords(self):
        """Function to get cursor's coordinates in appropriate coordinate system.
        
        Returns:
            Coordinates of the cursor relative to the center of the game object.
        """
        if not self.game_object.parent:
            return Vector2d.from_tuple(pg.mouse.get_pos()) - Vector2d.from_tuple(pg.display.get_window_size()) // 2 + GameObject.get_by_tag("camera").transform.position
        
        return self.game_object.parent.get_component(CoordinatesComponent).get_cursor_coords() - self.game_object.parent.transform.position
    
    def get_absolute_coords(self):
        """Function to get absolute game_object's coordinates.

        Returns:
            game_object's absolute coordinates.
        """
        if not self.game_object.parent:
            return self.game_object.transform.position
        
        return self.game_object.parent.get_component(CoordinatesComponent).get_absolute_coords() + self.game_object.transform.position
    

T = TypeVar("T")

class GameObject:
    """Base of this engine. Every object in the game must be implemented by creating a game object and adding necessary components.

    active      If false, this game object won't be displayed and updated. The same is true for it's childs.
    tag         Tag of the object. There can't be two game objects with the same tag.
    """
    components: list[Component]
    active: bool
    tag: str
    parent: "GameObject | None"
    transform: Transform
    surface: SurfaceComponent
    childs: list["GameObject"]

    tag_objects: dict[str, "GameObject"] = {}
    objects: list["GameObject"] = []

    def __init__(self, tag: str, root: bool = True, surf_size: Vector2d = Vector2d(800, 600)):
        self.components = []
        self.surface = SurfaceComponent(surf_size)
        self.add_component(self.surface)
        self.add_component(CoordinatesComponent())

        self.childs = []
        self.active = True
        self.tag = tag
        self.parent = None
        self.transform = Transform(self)

        if (self.tag in GameObject.tag_objects.keys()):
            raise Exception(f"Tried to create two object with same tag: {self.tag}")

        if root:
            GameObject.objects.append(self)
            
        GameObject.tag_objects[self.tag] = self

    def draw(self):
        center = self.get_component(CoordinatesComponent).get_absolute_coords()
        screen_tl = GameObject.get_by_tag("camera").transform.position - Vector2d.from_tuple(pg.display.get_window_size()) / 2
        screen_br = GameObject.get_by_tag("camera").transform.position + Vector2d.from_tuple(pg.display.get_window_size()) / 2
        obj_tl = center - Vector2d.from_tuple(self.surface.pg_surf.get_size()) / 2
        obj_br = center + Vector2d.from_tuple(self.surface.pg_surf.get_size()) / 2

        # Draw optimization. No need to draw the object if it is out of screen's bounds.
        if ((screen_br.x < obj_tl.x or screen_tl.x > obj_br.x) or (screen_br.y < obj_tl.y or screen_tl.y > obj_br.y)):
            return

        for component in self.components:
            component.draw()
        for child in self.childs:
            child.draw()
            child.surface.blit()

    def update(self):
        for i in self.components:
            i.update()
        for i in self.childs:
            i.update()

    def on_destroy(self):
        for i in self.childs:
            GameObject.destroy(i)

    def add_component(self, component: Component):
        component.init(self)
        self.components.append(component)

    def get_component(self, component: type[T]) -> T:
        for i in self.components:
            if isinstance(i, component):
                return i
        raise Exception(f"No such component: {component}")
    
    def contains_component(self, component: type[T]) -> T:
        for i in self.components:
            if isinstance(i, component):
                return True
        return False
    
    def add_child(self, child: "GameObject"):
        child.parent = self
        self.childs.append(child)

    def set_active(self, active: bool):
        self.active = active

    def clone(self) -> "GameObject":
        # TODO
        raise NotImplemented()
    
    def clear_surface(self):
        for i in self.childs:
            i.clear_surface()
        self.surface.pg_surf.fill((0, 0, 0, 0))

    @staticmethod
    def get_by_tag(tag: str) -> "GameObject":
        if not tag in GameObject.tag_objects.keys():
            raise KeyError(f"No object with tag: {tag}")
        return GameObject.tag_objects[tag]

    @staticmethod
    def destroy(obj: "GameObject"):
        obj.on_destroy()

    def __str__(self):
        return f"GameObject {self.tag}"



from ..game_object import GameObject, Component
from typing import Callable, Any
import pygame as pg


class KeybindComponent(Component):
    cmd: Callable[[GameObject, list[int]], Any]
    listen: list[int]
    listen_for_hold: bool
    on_press: bool
    previous: list[int]


    def __init__(self, listen: list[int], cmd: Callable[[GameObject, list[int]], Any], listen_for_hold: bool = False, on_press: bool = True):
        self.listen = listen
        self.listen_for_hold = listen_for_hold
        self.on_press = on_press
        self.cmd = cmd
    
    def update(self):
        keys = pg.key.get_pressed()
        listen_keys = []
        
        for i in self.listen:
            if keys[i]: listen_keys.append(i)

        active_keys = []
        if not self.listen_for_hold:
            for i in self.listen:
                if (self.on_press == (i in listen_keys) == (not (i in self.previous))):
                    active_keys.append(i)
            self.previous = listen_keys.copy()
        else:
            active_keys = listen_keys
        
        if len(active_keys) != 0:
            self.cmd(self.game_object, active_keys)


from ..game_object import Component
from ..globals import camera
import pygame as pg

class MovementComponent(Component):
    """Binds position of the camera to the position of the game object. Also implements basic camera movement.
    """
    speed: float

    def __init__(self, speed):
        super().__init__()
        self.speed = speed

    def update(self):
        pressed = pg.key.get_pressed()

        if pressed[pg.K_w]:
            self.game_object.transform.position.y -= self.speed
        if pressed[pg.K_s]:
            self.game_object.transform.position.y += self.speed
        if pressed[pg.K_a]:
            self.game_object.transform.position.x -= self.speed
        if pressed[pg.K_d]:
            self.game_object.transform.position.x += self.speed

        camera.transform.position = self.game_object.transform.position


from typing import TypeVar
import jsonpython

T = TypeVar("T")

def configclass(filename: str):
    """A decorator that transforms data from file to the decorated class.

    Args:
        filename    Name of the JSON file that contains data.

    Example:

    # config.json
    # {
    #   "lol": 1,
    #   "name": "sss"
    # }


    @config_class("config.json")
    class Config:
        lol: int
        name: str

    config = Config()
    print(config.name) # sss
    """

    def wrapper(cls: type[T]) -> type[T]:
        def load(self: T) -> None:
            obj = jsonpython.from_file(cls, filename)  # type: ignore
            self.__dict__.update(obj.__dict__)

        cls.__init__ = load

        return cls

    return wrapper



from .config import *
from .components.ui import *
from .components.keybind import *
from .components.movement import *
from .components.sprite import *
from .game_object import *
from .engine import *
from .tween import *
from .globals import *
from . import vmath

__all__ = [
    "Engine",
    "GameObject",
    "CoordinatesComponent",
    "Transform",
    "Component",
    "SurfaceComponent",
    "camera",
    "events",
    "Scene",
    "Tweens",
    "vmath",
    "KeybindComponent",
    "MovementComponent",
    "SpriteComponent",
    "LabelComponent",
    "ShapeComponent",
    "RectShapeComponent",
    "CircleShapeComponent",
    "ButtonComponent",
]




from .game_object import GameObject


class Scene:
    """Base class for scenes.
    
    name    Name of the scene.
    """
    name: str

    def __init__(self, name: str):
        self.name = name
        
    def load(self, data: any):
        """Loads the scene, creates necessary game objects and components.
        
        Args:
            data    Data that was given to the scene.
        """
        pass

    def destroy(self):
        for i in GameObject.objects:
            GameObject.destroy(i)

        GameObject.objects = []



import traceback
from .scene import Scene
from .game_object import GameObject
from .globals import events

import pygame as pg
import logging


class Engine:
    """Root class of the engine. Can't be instantiated more than once.

    display     Basically pygame.display.get_display()
    scene       Current loaded scene
    scenes      List of all registered scenes
    """
    instance = None

    scenes: list[Scene]
    scene: Scene
    logger: logging.Logger
    scene: Scene
    display: pg.Surface
    fps: int

    def __init__(self, app_name: str, fps: int, resolution: tuple[int, int] = (800, 600)):
        self.fps = fps
        self.scenes = []
        self.scene = None
        self.app_name = app_name
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("[%(levelname)s][%(name)s]: %(message)s"))
        self.logger.addHandler(handler)
        self.init(resolution)

    def __new__(cls, *args, **kwargs):
        if not cls.instance:
            cls.instance = super(Engine, cls).__new__(cls)
        return cls.instance

    def init(self, resolution: tuple[int, int] = (800, 600)):
        pg.init()
        pg.font.init()
        logging.basicConfig()
        self.display = pg.display.set_mode(resolution)
        pg.display.set_caption(self.app_name)
        logging.info("Engine initialized.")

    def register(self, scene: Scene):
        self.scenes.append(scene)

    async def load_scene(self, scene_name: str, data: any):
        scene = next((x for x in self.scenes if x.name == scene_name), None)

        if not scene:
            logging.error(f"Scene {scene_name} not found.")
            return
        
        if self.scene:
            scene.destroy()
            
        self.scene = scene
        scene.load(data)
        
        logging.info(f"Scene {scene.name} loaded")
        try:
            clock = pg.time.Clock()
            while True:
                await self.iteration()
                clock.tick(self.fps)
        except Exception as e:
            logging.critical(traceback.format_exc())
            exit(1)

    def event_processing(self, event: pg.event.Event):
        if event.type == pg.QUIT:
            logging.info("Quitting")
            exit()

    async def iteration(self):
        
        events = pg.event.get()
        
        for event in events:
            self.event_processing(event)
        
        for i in GameObject.objects:
            i.update()

        self.display.fill((0, 0, 0))
        for i in GameObject.objects:
            i.draw()
            i.surface.blit()

        pg.display.flip()

        for i in GameObject.objects:
            i.clear_surface()
            



import pygame
import os

from pygame_tools_tafh.vmath.vector import Vector2d

from ..game_object import Component

class SpriteComponent(Component):
    """A component to load sprites.

    path    Base path to all sprites.
    loaded  Cache for the sprites.
    """
    path: str = ''
    loaded: dict = {}

    def __init__(self, sprite_name: str, size: tuple[int, int]) -> None:
        super().__init__()

        if sprite_name in SpriteComponent.loaded.keys():
            self.texture = SpriteComponent.loaded[sprite_name]
        else:
            self.texture = pygame.image.load(os.path.join(SpriteComponent.path, sprite_name)).convert_alpha()
            SpriteComponent.loaded[sprite_name] = self.texture

        self.size = size
        self.opacity = 255

    @staticmethod
    def set_path(path: str):
        SpriteComponent.path = path

    def draw(self):
        display = self.game_object.surface.pg_surf
        self.texture.set_alpha(self.opacity)
        blit_image = self.texture

        cropped = pygame.Surface(self.size)
        cropped.blit(blit_image, (0, 0))

        angle = self.game_object.transform.angle.get()
        scale = self.game_object.transform.scale

        if angle != 0:
            cropped = pygame.transform.rotate(cropped, angle)

        if scale != 1:
            cropped = pygame.transform.scale_by(cropped, scale)

        rect = cropped.get_rect(center=(Vector2d.from_tuple(display.get_size()) / 2).as_tuple())

        display.blit(cropped, rect)
    


from .vector import *


def is_in_box(vec1: Vector2d, other1: "Vector2d", other2: "Vector2d") -> bool:
    """Checks if the position is inside of the box.

    Args:
        vec1    Position to check.
        other1  Position of the one vertex of the box.
        other2  Position of the diametrically opposed vertex of the box.

    Returns:
        True if the position is inside of the box, False otherwise.
    """
    return ((vec1 - other1) * (other2 - other1)).get_quarter() == 1 and (
            (other2 - other1) * (other2 - other1) - (vec1 - other1) * (vec1 - other1)).get_quarter() == 1

def complex_multiply(vec1: Vector2d, other: Vector2d) -> Vector2d:
    return Vector2d(vec1.x * other.x - vec1.y * other.y, vec1.y * other.x + vec1.x * other.y)


from typing import Callable
from .angle import *

class Vector2d:
    """Class to represent a pair of floats.

    x   X component of the vector
    y   Y component of the vector
    """

    x: float
    y: float

    def __init__(self, a: float = 0, b: float = 0):
        self.x = a
        self.y = b

    @staticmethod
    def from_tuple(tpl: tuple[float, float]) -> "Vector2d":
        return Vector2d(tpl[0], tpl[1])

    def as_tuple(self) -> tuple[float, float]:
        return self.x, self.y

    def mod(self) -> float:
        """Returns length of the vector.
        """
        return sqrt((self.x ** 2) + (self.y ** 2))

    def intx(self) -> int:
        return int(self.x)

    def inty(self) -> int:
        return int(self.y)

    def to_angle(self) -> "Angle":
        return Angle(atan2(-self.y, self.x))

    def get_quarter(self) -> int:
        if self.x >= 0 and self.y >= 0:
            return 1
        elif self.x <= 0 and self.y <= 0:
            return 3
        elif self.x < 0:
            return 2
        elif self.y < 0:
            return 4
        
    def operation(self, other: "Vector2d", operation: Callable[[float, float], float]) -> "Vector2d":
        return Vector2d(operation(self.x, other.x), operation(self.y, other.y))
    
    def __add__(self, other: "Vector2d") -> "Vector2d":
        return Vector2d(self.x + other.x, self.y + other.y)

    def __sub__(self, other: "Vector2d") -> "Vector2d":
        return Vector2d(self.x - other.x, self.y - other.y)

    def __mul__(self, other: "float | Vector2d") -> "Vector2d":
        if isinstance(other, Vector2d):
            return Vector2d(self.x * other.x, self.y * other.y)
        else:
            return Vector2d(self.x * other, self.y * other)

    def __floordiv__(self, other: float):
        return Vector2d(self.x // other, self.y // other)

    def __truediv__(self, other: float) -> "Vector2d":
        return Vector2d(self.x / other, self.y / other)

    def __mod__(self, other: float) -> "Vector2d":
        return Vector2d(self.x % other, self.y % other)

    def __repr__(self) -> str:  # for debugging
        return f"<{self.x}, {self.y}>"

    def __eq__(self, other: "Vector2d") -> bool:
        return self.x == other.x and self.y == other.y

    def __ne__(self, other: "Vector2d") -> bool:
        return self.x != other.x or self.y != other.y



from math import *

from .vector import Vector2d


class Angle:
    """Class that represent angles in radians.

    angle   Angle in radians.
    """

    angle: float

    def __init__(self, angle: float = 0) -> None:
        self.angle = angle
        self.angle %= (2 * pi)

    def set(self, angle: float, is_degree: bool = False):
        if is_degree:
            angle = angle * pi / 180
        self.angle = angle
        self.angle %= (2 * pi)

    def get(self, is_degree: bool = False):
        if is_degree:
            return self.angle * 180 / pi
        return self.angle

    def to_vector(self) -> Vector2d:
        """Breaks angle into it's cosine and sine components.

        Returns:
            Vectored value of the angle.
        """
        return Vector2d(cos(self.angle), sin(self.angle))

    def __add__(self, other: "Angle") -> "Angle":
        return Angle(self.get() + other.get())

    def __sub__(self, other: "Angle") -> "Angle":
        return Angle(self.get() - other.get())

    def __repr__(self) -> str:
        return str(self.angle)

    def __float__(self) -> float:
        return self.angle


class Modx:
    """Value by the certain modulus.

    value   Value.
    x       Modulus.
    """

    value: int
    x: int

    def __init__(self, value: int = 0, x: int = 4) -> None:
        self.value = value % x
        self.x = x

    def __add__(self, other: "Modx") -> "Modx":
        return Modx((self.value + other.value) % self.x)

    def __sub__(self, other: "Modx") -> "Modx":
        return Modx((self.value - other.value) % self.x)

    def __repr__(self) -> str:
        return self.value.__repr__()

    def __eq__(self, other: "Modx") -> bool:
        return self.value == other.value

    def __ne__(self, other: "Modx") -> bool:
        return self.value != other.value


from .angle import *
from .modx import *
from .vector import *

class Direction(Modx):
    def __init__(self, value: int = 0) -> None:
        super().__init__(value)

    def to_angle(self) -> Angle:
        return Angle(self.value * pi / 2)

    @staticmethod
    def from_angle(ang: Angle) -> "Direction":
        return Direction((ang.get() + pi / 4) // (pi / 2))

    def to_vector(self) -> Vector2d:
        return Directions.AsVector2D[self.value]


class Directions:
    RIGHT = Direction(0)
    UP = Direction(1)
    LEFT = Direction(2)
    DOWN = Direction(3)

    AsVector2D = (
        Vector2d(1, 0),
        Vector2d(0, -1),
        Vector2d(-1, 0),
        Vector2d(0, 1)
    )


from typing import Callable
import socket

SEPARATOR = bytes([0x00])

class Serializable:

    def serialize(self) -> dict[str, str | int]:
        return {}

    def to_bytes(self) -> bytes:
        return serialize(self.serialize())

def serialize(dc: dict[str, str | int]) -> bytes:
    data = bytes()
    for key, value in dc.items():
        data += key.encode()
        data += SEPARATOR
        if isinstance(value, str):
            data += bytes([1])
            data += value.encode()
            data += bytes([0])
        elif isinstance(value, int):
            data += bytes([2])
            data += value.to_bytes(4, "big")
        else:
            raise Exception("Invalid type")

    data += bytes([0, 0])
    return data

class PlayerConn:
    conn: socket.socket
    name: str
    _listeners: dict[str, list[Callable]]

    def __init__(self, conn) -> None:
        self.conn = conn
        self._listeners = {}

    def on(self, event: str, listener: Callable):
        if event not in self._listeners.keys():
            self._listeners[event] = []
        self._listeners[event].append(listener)

    def emit(self, event: str, data: Serializable):
        self.conn.send(event.encode())
        self.conn.send(SEPARATOR)
        self.conn.send(data.to_bytes())

    async def load(self):
        event = self.read_string()
        data = self.read_object()

        if event in self._listeners.keys():
            for listener in self._listeners[event]:
                listener(data)

    def load_name(self):
        name = self.read_string()
        self.name = name
        return name
    
    def read_object(self):
        obj = {}
        while 1:
            key = self.read_string()
            tp = int.from_bytes(self.conn.recv(1), "big")

            match tp:
                case 0:
                    break
                case 1:
                    obj[key] = self.read_string()
                case 2:
                    obj[key] = self.read_int()
                case _:
                    break
        return obj

    def read_string(self):
        bts = bytes()
        while 1:
            bt = self.conn.recv(1)
            if bt == SEPARATOR:
                break
            bts += bt
        return bts.decode()
    
    def read_int(self):
        bts = self.conn.recv(4)
        return int.from_bytes(bts, "big")